#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <c10/cuda/CUDAGuard.h>

#define WARP_SIZE 32
#define TILE_SIZE 4

__global__ void hybrid_optimized_kernel(const float* __restrict__ x,
                                         const float* __restrict__ weight,
                                         const float* __restrict__ bias,
                                         float* __restrict__ out,
                                         int batch_size,
                                         int in_features,
                                         int out_features) {
    int batch_idx = blockIdx.x;
    int warp_id = threadIdx.x / WARP_SIZE;
    int lane_id = threadIdx.x % WARP_SIZE;
    int warps_per_block = blockDim.x / WARP_SIZE;
    int base_out_group = blockIdx.y * (warps_per_block * TILE_SIZE);
    int out_base = base_out_group + warp_id * TILE_SIZE;

    if (out_base >= out_features) return;

    float sums[TILE_SIZE] = {0.0f, 0.0f, 0.0f, 0.0f};
    const float* x_row = x + batch_idx * in_features;

    for (int tile = 0; tile < TILE_SIZE; tile++) {
        int current_out = out_base + tile;
        if (current_out < out_features) {
            const float* w_row = weight + current_out * in_features;
            int vec_count = in_features / 4;
            const float4* x_vec = reinterpret_cast<const float4*>(x_row);
            const float4* w_vec = reinterpret_cast<const float4*>(w_row);

            #pragma unroll
            for (int i = lane_id; i < vec_count; i += WARP_SIZE) {
                float4 xv = __ldg(x_vec + i);
                float4 wv = __ldg(w_vec + i);
                sums[tile] += xv.x * wv.x + xv.y * wv.y + xv.z * wv.z + xv.w * wv.w;
            }

            int rem_start = vec_count * 4;
            for (int i = rem_start + lane_id; i < in_features; i += WARP_SIZE) {
                sums[tile] += __ldg(x_row + i) * __ldg(w_row + i);
            }
        }
    }

    for (int tile = 0; tile < TILE_SIZE; tile++) {
        for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
            sums[tile] += __shfl_down_sync(0xffffffff, sums[tile], offset);
        }
    }

    if (lane_id == 0) {
        for (int tile = 0; tile < TILE_SIZE; tile++) {
            int current_out = out_base + tile;
            if (current_out < out_features) {
                float result = sums[tile] + __ldg(bias + current_out);
                out[batch_idx * out_features + current_out] = (result > 0.0f) ? result : 0.0f;
            }
        }
    }
}

torch::Tensor hybrid_optimized_forward(torch::Tensor x,
                                       torch::Tensor weight,
                                       torch::Tensor bias) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    int batch_size = x.size(0);
    int in_features = x.size(1);
    int out_features = weight.size(0);

    auto out = torch::empty({batch_size, out_features}, x.options());

    int warps_per_block = 8;
    int threads_per_block = warps_per_block * WARP_SIZE;
    int blocks_y = (out_features + (warps_per_block * TILE_SIZE) - 1) / (warps_per_block * TILE_SIZE);

    dim3 grid(batch_size, blocks_y);
    dim3 block(threads_per_block);

    cudaStream_t stream = c10::cuda::getCurrentCUDAStream();
    hybrid_optimized_kernel<<<grid, block, 0, stream>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        batch_size,
        in_features,
        out_features
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &hybrid_optimized_forward, "Optimized GEMM with bias and ReLU using hybrid approach (CUDA)");
}
